@@ -16,6 +16,7 @@ from mcp.types import Resource, Tool
 
 from .exceptions import MissingConfigError
 from ..config import MCPConfig, ServerConfig
+from .utils import exponential_retry
 from pprint import pformat
 
 # Configure the logger
@@ -46,6 +47,8 @@ DEFAULT_MODEL_MAP = {
     }
 }
 
+class LLMStreamError(Exception):
+    pass
 
 class MCPClient:
     """Client for interacting with LLMs and MCP servers.
@@ -531,17 +534,46 @@ class MCPClient:
             current_tool_args = ""
             tool_call_complete = False
             # Start streaming response
-            start_reason = True
-            response_stream = await litellm.acompletion(
-                model=self.model,
-                messages=messages,
-                max_tokens=self.max_tokens,
-                temperature=temperature,
-                tools=tools,
-                tool_choice="auto",
-                stream=True,
-                api_base=self.base_url,
+            start_reason = True  # flags when llm is outputting reasoning/thinking
+            reason_complete = True # flags when the thinking part is done
+            @exponential_retry(
+                max_retries=10,
+                base_delay=62.0,
+                max_delay=600.0,
+                jitter=True,
+                retryable_exceptions=(
+                        litellm.InternalServerError,
+                        ConnectionError,
+                        TimeoutError,
+                        litellm.RateLimitError,
+                        litellm.ContextWindowExceededError,
+                ),
+                logger=logger,
             )
+            async def get_stream():
+                try:
+                    return await litellm.acompletion(
+                        model=self.model,
+                        messages=messages,
+                        max_tokens=self.max_tokens,
+                        temperature=temperature,
+                        tools=tools,
+                        tool_choice="auto",
+                        stream=True,
+                        api_base=self.base_url,
+                        api_key=self.api_key
+                    )
+                except litellm.RateLimitError as ex:
+                    logger.error(f'got rate limited will retry after a minute: {ex}')
+                    raise
+                except Exception as ex:
+                    logger.error(f"unhandled exception: {ex}")
+                    raise
+            
+            response_stream = await get_stream()   
+            if not response_stream:
+                logger.error(f"problem getting stream from llm")
+                raise LLMStreamError()
 
             async for chunk in response_stream:
                 # Check for tool calls
@@ -612,11 +644,17 @@ class MCPClient:
                     yield chunk.choices[0]
                 delta = chunk.choices[0].delta
                 if hasattr(delta, 'content') and delta.content and not tool_call_complete:
-                    yield delta.content
+                    if not reason_complete:
+                        reason_complete = True
+                        start_reason = False
+                        yield "</think>\n" + delta.content
+                    else:
+                        yield delta.content
                 elif hasattr(delta, 'reasoning_content') and delta.reasoning_content:
                     if start_reason:
                         start_reason = False
-                        yield  "<think>" + delta.reasoning_content
+                        reason_complete = False
+                        yield  "<think>\n" + delta.reasoning_content
                     else:
                         yield delta.reasoning_content
                 elif hasattr(delta, 'finish_reason') and delta.finish_reason:
